import { __awaiter } from 'tslib';
import { inject } from '@angular/core';
import { AngularFirestore } from '@angular/fire/firestore';
import { withTransaction, updateEntities, arrayAdd, arrayUpsert, arrayRemove, arrayUpdate } from '@datorama/akita';
import { isObservable, of, combineLatest, race } from 'rxjs';
import { map, switchMap, tap, distinctUntilChanged, filter, shareReplay, finalize } from 'rxjs/operators';
import { Router } from '@angular/router';
import { RouterQuery } from '@datorama/akita-ng-router-store';
import { AngularFireAuth } from '@angular/fire/auth';
import { auth } from 'firebase';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// Helper to retrieve the id and path of a document in the collection
/**
 * @param {?} options
 * @param {?=} collectionPath
 * @return {?}
 */
function getIdAndPath(options, collectionPath) {
    /** @type {?} */
    let path = '';
    /** @type {?} */
    let id = '';
    if (options['id']) {
        if (!collectionPath) {
            throw new Error('You should provide the colletion path with the id');
        }
        id = options['id'];
        path = `${collectionPath}/${id}`;
    }
    else if (options['path']) {
        path = options['path'];
        /** @type {?} */
        const part = path.split('/');
        if (path.length % 2 !== 0) {
            throw new Error(`Path ${path} doesn't look like a Firestore's document path`);
        }
        id = part[part.length - 1];
    }
    else {
        throw new Error(`You should provide either an "id" OR a "path".`);
    }
    return { id, path };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// Credit to @arielgueta https://dev.to/arielgueta/getting-started-with-akita-and-firebase-3pe2
/**
 * @template S
 * @this {?}
 * @param {?} actions
 * @return {?}
 */
function syncFromAction(actions) {
    this.store.setLoading(false);
    if (actions.length === 0) {
        return;
    }
    for (const action of actions) {
        /** @type {?} */
        const id = (/** @type {?} */ (action.payload.doc.id));
        /** @type {?} */
        const entity = action.payload.doc.data();
        switch (action.type) {
            case 'added': {
                this.store.upsert(id, Object.assign({ [this.idKey]: id }, entity));
                break;
            }
            case 'removed': {
                this.store.remove(id);
                break;
            }
            case 'modified': {
                this.store.update(id, entity);
            }
        }
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template S
 */
class CollectionService {
    /**
     * @param {?} store
     * @param {?=} collectionPath
     */
    constructor(store, collectionPath) {
        this.store = store;
        this.collectionPath = collectionPath;
        if (!this.constructor['path'] && !this.collectionPath) {
            throw new Error('You should provide a path to the collection');
        }
        try {
            this.db = inject(AngularFirestore);
        }
        catch (err) {
            throw new Error('CollectionService requires AngularFirestore.');
        }
    }
    /**
     * @return {?}
     */
    get idKey() {
        return this.constructor['idKey'] || this.store.idKey;
    }
    /**
     * The path to the collection in Firestore
     * @return {?}
     */
    get path() {
        return this.constructor['path'] || this.collectionPath;
    }
    /**
     * A snapshot of the path
     * @return {?}
     */
    get currentPath() {
        if (isObservable(this.path)) {
            throw new Error('Cannot get a snapshot of the path if it is an Observable');
        }
        return this.path;
    }
    /**
     * An observable version of the path
     * @return {?}
     */
    get path$() {
        return isObservable(this.path) ? this.path : of(this.path);
    }
    /**
     * The Angular Fire collection
     * \@notice If path is an observable, it becomes an observable.
     * @return {?}
     */
    get collection() {
        return this.db.collection(this.currentPath);
    }
    /**
     * Preformat the document before updating Firestore
     * @protected
     * @template E
     * @param {?} document
     * @return {?}
     */
    preFormat(document) {
        return document;
    }
    /**
     * The config given by the `CollectonConfig`
     * @return {?}
     */
    get config() {
        return {
            path: this.constructor['path'],
            idKey: this.constructor['idKey']
        };
    }
    /**
     * @param {?=} pathOrQuery
     * @param {?=} queryFn
     * @return {?}
     */
    syncCollection(pathOrQuery = this.path, queryFn) {
        /** @type {?} */
        let path;
        if (isObservable(pathOrQuery)) {
            path = pathOrQuery;
        }
        else if (typeof pathOrQuery === 'function') {
            queryFn = pathOrQuery;
            path = this.path$;
        }
        else {
            path = this.path$;
        }
        // If there is no entity yet set loading
        if (!this.store._value().ids || this.store._value().ids.length === 0) {
            this.store.setLoading(true);
        }
        // Start Listening
        return path.pipe(map((/**
         * @param {?} collectionPath
         * @return {?}
         */
        collectionPath => this.db.collection(collectionPath, queryFn))), switchMap((/**
         * @param {?} collection
         * @return {?}
         */
        collection => collection.stateChanges())), withTransaction(syncFromAction.bind(this)));
    }
    /**
     * @param {?=} idOrQuery
     * @param {?=} queryGroupFn
     * @return {?}
     */
    syncCollectionGroup(idOrQuery, queryGroupFn) {
        /** @type {?} */
        const path = (typeof idOrQuery === 'string') ? idOrQuery : this.currentPath;
        /** @type {?} */
        const query = typeof idOrQuery === 'function' ? idOrQuery : queryGroupFn;
        /** @type {?} */
        const collectionId = path.split('/').pop();
        return this.db.collectionGroup(collectionId, query)
            .stateChanges()
            .pipe(withTransaction(syncFromAction.bind(this)));
    }
    /**
     * Sync the store with several documents
     * @param {?} ids An array of ids
     * @return {?}
     */
    syncManyDocs(ids) {
        /** @type {?} */
        const syncs = ids.map((/**
         * @param {?} id
         * @return {?}
         */
        id => this.path$.pipe(map((/**
         * @param {?} collectionPath
         * @return {?}
         */
        collectionPath => getIdAndPath({ id }, collectionPath))), switchMap((/**
         * @param {?} __0
         * @return {?}
         */
        ({ path }) => this.db.doc(path).valueChanges())), map((/**
         * @param {?} doc
         * @return {?}
         */
        doc => ((/** @type {?} */ (Object.assign({ [this.idKey]: id }, doc)))))))));
        return combineLatest(syncs).pipe(tap((/**
         * @param {?} entities
         * @return {?}
         */
        entities => this.store.upsertMany(entities))));
    }
    /**
     * Stay in sync with one document
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} options An object with EITHER `id` OR `path`.
     * @return {?}
     */
    syncDoc(options) {
        /** @type {?} */
        let id;
        return this.path$.pipe(map((/**
         * @param {?} collectionPath
         * @return {?}
         */
        collectionPath => getIdAndPath(options, collectionPath))), tap((/**
         * @param {?} data
         * @return {?}
         */
        data => {
            id = (/** @type {?} */ (data.id));
            if (!this.store._value().ids.includes(id)) {
                this.store.setLoading(true);
            }
        })), switchMap((/**
         * @param {?} __0
         * @return {?}
         */
        ({ path }) => this.db.doc(path).valueChanges())), map((/**
         * @param {?} entity
         * @return {?}
         */
        (entity) => {
            /** @type {?} */
            const data = Object.assign({ [this.idKey]: id }, entity);
            this.store.upsert(id, data);
            this.store.setLoading(false);
            return data;
        })));
    }
    /**
     * @param {?} options
     * @return {?}
     */
    syncActive(options) {
        if (Array.isArray(options)) {
            return this.syncManyDocs(options).pipe(tap((/**
             * @param {?} _
             * @return {?}
             */
            _ => this.store.setActive((/** @type {?} */ (options))))));
        }
        else {
            return this.syncDoc(options).pipe(tap((/**
             * @param {?} entity
             * @return {?}
             */
            entity => this.store.setActive(entity[this.idKey]))));
        }
    }
    /**
     * @param {?=} idOrQuery
     * @return {?}
     */
    getValue(idOrQuery) {
        return __awaiter(this, void 0, void 0, function* () {
            // If path targets a collection ( odd number of segments after the split )
            if (typeof idOrQuery === 'string') {
                /** @type {?} */
                const snapshot = yield this.db.doc(`${this.currentPath}/${idOrQuery}`).ref.get();
                return (/** @type {?} */ (snapshot.data()));
            }
            else {
                /** @type {?} */
                const snapshot = yield this.db.collection(this.currentPath, idOrQuery).ref.get();
                return snapshot.docs.map((/**
                 * @param {?} doc
                 * @return {?}
                 */
                doc => (/** @type {?} */ (doc.data()))));
            }
        });
    }
    /**
     * Add a document or a list of document to Firestore
     * @param {?} documents
     * @param {?=} options
     * @return {?}
     */
    add(documents, options) {
        return __awaiter(this, void 0, void 0, function* () {
            // Add id to the document
            /** @type {?} */
            const addId = (/**
             * @param {?} doc
             * @return {?}
             */
            (doc) => (Object.assign({}, doc, { [this.idKey]: doc[this.idKey] || this.db.createId() })));
            // Add the list of document with the right atomic operation
            /** @type {?} */
            const addDocuments = (/**
             * @param {?} operation
             * @param {?} writeOptions
             * @return {?}
             */
            (operation, writeOptions) => {
                /** @type {?} */
                const docs = Array.isArray(documents) ? documents : [documents];
                /** @type {?} */
                const operations = docs.map((/**
                 * @param {?} document
                 * @return {?}
                 */
                document => {
                    /** @type {?} */
                    const doc = addId(this.preFormat(document));
                    const { ref } = this.db.doc(`${this.currentPath}/${doc[this.idKey]}`);
                    if (this.onCreate) {
                        this.onCreate(doc, writeOptions);
                    }
                    return operation(ref, doc);
                }));
                return Promise.all(operations);
            });
            if (options && options.write) {
                return addDocuments((/**
                 * @param {?} ref
                 * @param {?} doc
                 * @return {?}
                 */
                (ref, doc) => options.write.set(ref, doc)), options);
            }
            else {
                /** @type {?} */
                const batch = this.db.firestore.batch();
                yield addDocuments((/**
                 * @param {?} ref
                 * @param {?} doc
                 * @return {?}
                 */
                (ref, doc) => batch.set(ref, doc)), {
                    write: batch,
                    ctx: options ? options.ctx : undefined
                });
                return batch.commit();
            }
        });
    }
    /**
     * Remove one or several document from Firestore
     * @param {?} id A unique or list of id representing the document
     * @param {?=} options
     * @return {?}
     */
    remove(id, options) {
        return __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            const removeDocuments = (/**
             * @param {?} operation
             * @param {?} writeOptions
             * @return {?}
             */
            (operation, writeOptions) => {
                /** @type {?} */
                const ids = Array.isArray(id) ? id : [id];
                /** @type {?} */
                const operations = ids.map((/**
                 * @param {?} docId
                 * @return {?}
                 */
                (docId) => __awaiter(this, void 0, void 0, function* () {
                    const { ref } = this.db.doc(`${this.currentPath}/${docId}`);
                    if (this.onDelete) {
                        yield this.onDelete(docId, writeOptions);
                    }
                    return operation(ref);
                })));
                return Promise.all(operations);
            });
            if (options && options.write) {
                return removeDocuments((/**
                 * @param {?} ref
                 * @return {?}
                 */
                (ref) => options.write.delete(ref)), options);
            }
            else {
                /** @type {?} */
                const batch = this.db.firestore.batch();
                yield removeDocuments((/**
                 * @param {?} ref
                 * @return {?}
                 */
                (ref) => batch.delete(ref)), {
                    write: batch,
                    ctx: options ? options.ctx : undefined
                });
                return batch.commit();
            }
        });
    }
    /**
     * @param {?} idsOrFn
     * @param {?=} stateFnOrWrite
     * @param {?=} options
     * @return {?}
     */
    update(idsOrFn, stateFnOrWrite, options) {
        return __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            const isEntity = (/**
             * @param {?} value
             * @return {?}
             */
            (value) => {
                return typeof value === 'object' && idsOrFn[this.idKey];
            });
            // Set variables
            /** @type {?} */
            let ids = [];
            /** @type {?} */
            let newStateOrFn = (/** @type {?} */ (stateFnOrWrite));
            if (isEntity(idsOrFn)) {
                ids = [idsOrFn[this.idKey]];
                options = (/** @type {?} */ (stateFnOrWrite));
                newStateOrFn = idsOrFn;
            }
            else if (typeof idsOrFn === 'function') {
                /** @type {?} */
                const state = this.store._value();
                ids = (/** @type {?} */ (state.ids.filter((/**
                 * @param {?} id
                 * @return {?}
                 */
                id => idsOrFn(state.entities[id])))));
            }
            else {
                ids = Array.isArray(idsOrFn) ? idsOrFn : [idsOrFn];
            }
            if (!Array.isArray(ids) || ids.length === 0) {
                return;
            }
            // Create updates
            /** @type {?} */
            const updates = updateEntities({
                state: this.store._value(),
                ids,
                idKey: this.idKey,
                newStateOrFn,
                preUpdateEntity: this.store.akitaPreUpdateEntity
            }).entities;
            // Update the list of document with the right atomic operation
            /** @type {?} */
            const updateDocuments = (/**
             * @param {?} operation
             * @param {?} writeOptions
             * @return {?}
             */
            (operation, writeOptions) => {
                /** @type {?} */
                const operations = Object.keys(updates).map((/**
                 * @param {?} key
                 * @return {?}
                 */
                (key) => __awaiter(this, void 0, void 0, function* () {
                    /** @type {?} */
                    const doc = this.preFormat(updates[key]);
                    const { ref } = this.db.doc(`${this.currentPath}/${doc[this.idKey]}`);
                    if (this.onUpdate) {
                        yield this.onUpdate(doc, writeOptions);
                    }
                    return operation(ref, doc);
                })));
                return Promise.all(operations);
            });
            if (options && options.write) {
                return updateDocuments((/**
                 * @param {?} ref
                 * @param {?} doc
                 * @return {?}
                 */
                (ref, doc) => options.write.update(ref, doc)), options);
            }
            else {
                /** @type {?} */
                const batch = this.db.firestore.batch();
                yield updateDocuments((/**
                 * @param {?} ref
                 * @param {?} doc
                 * @return {?}
                 */
                (ref, doc) => batch.update(ref, doc)), {
                    write: batch,
                    ctx: options ? options.ctx : undefined
                });
                return batch.commit();
            }
        });
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Set the configuration for the collection service
 * @param {?=} options
 * @return {?}
 */
function CollectionConfig(options = {}) {
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    (constructor) => {
        Object.keys(options).forEach((/**
         * @param {?} key
         * @return {?}
         */
        key => constructor[key] = options[key]));
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} data
 * @return {?}
 */
function CollectionGuardConfig(data) {
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    constructor => {
        Object.keys(data).forEach((/**
         * @param {?} key
         * @return {?}
         */
        key => (constructor[key] = data[key])));
    });
}
/**
 * @template S
 */
class CollectionGuard {
    /**
     * @param {?} service
     */
    constructor(service) {
        this.service = service;
        try {
            this.router = inject(Router);
        }
        catch (err) {
            throw new Error('CollectionGuard requires RouterModule to be imported');
        }
    }
    // Can be override by the extended class
    /**
     * Should the guard wait for connection to Firestore to complete
     * @protected
     * @return {?}
     */
    get awaitSync() {
        return this.constructor['awaitSync'] || false;
    }
    // Can be override by the extended class
    /**
     * A queryFn to execute for sync
     * @protected
     * @return {?}
     */
    get queryFn() {
        return this.constructor['queryFn'];
    }
    // Can be override by the extended class
    /**
     * The route to redirect to if you sync failed
     * @protected
     * @return {?}
     */
    get redirect() {
        return this.constructor['redirect'];
    }
    // Can be override by the extended class
    /**
     * The method to subscribe to while route is active
     * @protected
     * @param {?} next
     * @return {?}
     */
    sync(next) {
        const { queryFn = this.queryFn } = (/** @type {?} */ (next.data));
        return this.service.syncCollection(queryFn);
    }
    /**
     * @param {?} next
     * @return {?}
     */
    canActivate(next) {
        const { redirect = this.redirect, awaitSync = this.awaitSync } = (/** @type {?} */ (next.data));
        return new Promise((/**
         * @param {?} res
         * @param {?} rej
         * @return {?}
         */
        (res, rej) => {
            if (awaitSync) {
                this.subscription = this.sync(next).subscribe({
                    next: (/**
                     * @param {?} result
                     * @return {?}
                     */
                    (result) => {
                        switch (typeof result) {
                            case 'string': {
                                this.subscription.unsubscribe();
                                return res(this.router.parseUrl(result));
                            }
                            case 'boolean':
                                return res(result);
                            default:
                                return res(true);
                        }
                    }),
                    error: (/**
                     * @param {?} err
                     * @return {?}
                     */
                    (err) => {
                        this.subscription.unsubscribe();
                        res(this.router.parseUrl(redirect || ''));
                        throw new Error(err);
                    })
                });
            }
            else {
                this.subscription = this.sync(next).subscribe();
                res(true);
            }
        }));
    }
    /**
     * @return {?}
     */
    canDeactivate() {
        if (this.subscription) {
            this.subscription.unsubscribe();
        }
        return true;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Get the params from a path
 * @param {?} path
 * @return {?}
 */
function getPathParams(path) {
    return path.split('/')
        .filter((/**
     * @param {?} segment
     * @return {?}
     */
    segment => segment.charAt(0) === ':'))
        .map((/**
     * @param {?} segment
     * @return {?}
     */
    segment => segment.substr(1)));
}
/**
 * Transform a path based on the params
 * \@example pathWithParams('movies/:movieId/stakeholder/:shId', { movieId, shId })
 * @param {?} path The path with params starting with "/:"
 * @param {?} params A map of id params
 * @return {?}
 */
function pathWithParams(path, params) {
    return path.split('/').map((/**
     * @param {?} segment
     * @return {?}
     */
    segment => {
        if (segment.charAt(0) === ':') {
            /** @type {?} */
            const key = segment.substr(1);
            if (!params[key]) {
                throw new Error(`Required parameter ${key} from ${path} doesn't exist in params ${JSON.stringify(params)}`);
            }
            return params[key];
        }
        else {
            return segment;
        }
    })).join('/');
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template S
 */
class SubcollectionService extends CollectionService {
    /**
     * @param {?} store
     */
    constructor(store) {
        super(store);
        this.store = store;
        try {
            this.routerQuery = inject(RouterQuery);
        }
        catch (err) {
            throw new Error('SubCollectionService requires a "AkitaNgRouterStoreModule" from Akita.');
        }
    }
    /**
     * @return {?}
     */
    get path() {
        /** @type {?} */
        const pathParams = getPathParams(this.constructor['path']);
        return this.routerQuery.selectParams().pipe(
        // Don't trigger change if params needed (and only them) haven't changed
        distinctUntilChanged((/**
         * @param {?} old
         * @param {?} next
         * @return {?}
         */
        (old, next) => {
            return !pathParams.find((/**
             * @param {?} param
             * @return {?}
             */
            param => old[param] !== next[param]));
        })), 
        // Need to filter because changes in params comes before canDeactivate
        filter((/**
         * @param {?} params
         * @return {?}
         */
        params => pathParams.every((/**
         * @param {?} param
         * @return {?}
         */
        param => !!params[param])))), map((/**
         * @param {?} params
         * @return {?}
         */
        params => pathWithParams(this.constructor['path'], params))), tap((/**
         * @param {?} path
         * @return {?}
         */
        path => {
            if (path !== this.pathToCollection) {
                this.store.reset();
                this.pathToCollection = path;
            }
        })), shareReplay(1));
    }
    /**
     * @return {?}
     */
    get currentPath() {
        /** @type {?} */
        const params = this.routerQuery.getParams();
        return pathWithParams(this.constructor['path'], params);
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @abstract
 * @template S
 */
class CollectionGroupService {
    /**
     * @param {?} store
     */
    constructor(store) {
        this.store = store;
        try {
            this.db = inject(AngularFirestore);
        }
        catch (err) {
            throw new Error('CollectionGroupService requires AngularFirestore.');
        }
    }
    /**
     * @return {?}
     */
    get idKey() {
        return this.store.idKey;
    }
    /**
     * Sync the collection group with the store
     * @param {?=} queryGroupFn
     * @return {?}
     */
    syncCollection(queryGroupFn) {
        return this.db.collectionGroup(this.collectionId, queryGroupFn).stateChanges()
            .pipe(withTransaction(syncFromAction.bind(this)));
    }
    /**
     * Return a snapshot of the collection group
     * @param {?=} queryGroupFn
     * @return {?}
     */
    getValue(queryGroupFn) {
        return __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            const snapshot = yield this.db.collectionGroup(this.collectionId, queryGroupFn).get().toPromise();
            return snapshot.docs.map((/**
             * @param {?} doc
             * @return {?}
             */
            doc => (/** @type {?} */ (doc.data()))));
        });
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const fireAuthProviders = (/** @type {?} */ (['github', 'google', 'microsoft', 'facebook', 'twitter', 'email']));
/**
 * @param {?} provider
 * @return {?}
 */
function isFireAuthProvider(provider) {
    return fireAuthProviders.includes((/** @type {?} */ (provider)));
}
/**
 * @param {?} provider
 * @return {?}
 */
function getAuthProvider(provider) {
    switch (provider) {
        case 'email': return new auth.EmailAuthProvider();
        case 'facebook': return new auth.FacebookAuthProvider();
        case 'github': return new auth.GithubAuthProvider();
        case 'google': return new auth.GoogleAuthProvider();
        case 'microsoft': return new auth.OAuthProvider('microsoft.com');
        case 'twitter': return new auth.TwitterAuthProvider();
    }
}
/**
 * @template S
 */
class FireAuthService {
    /**
     * @param {?} store
     */
    constructor(store) {
        this.store = store;
        this.collectionPath = 'users';
        this.db = inject(AngularFirestore);
        this.fireAuth = inject(AngularFireAuth);
        this.collection = this.db.collection(this.path);
    }
    /**
     * Can be overrided
     * @protected
     * @param {?} user
     * @return {?}
     */
    selectProfile(user) {
        return this.collection.doc(user.uid).valueChanges();
    }
    /**
     * Can be overrided
     * @protected
     * @param {?} user
     * @return {?}
     */
    selectRoles(user) {
        return user.getIdTokenResult().then((/**
         * @param {?} __0
         * @return {?}
         */
        ({ claims }) => (/** @type {?} */ (claims))));
    }
    /**
     * Should be override
     * @protected
     * @param {?} user
     * @return {?}
     */
    createProfile(user) {
        return (/** @type {?} */ ({
            photoURL: user.photoURL,
            displayName: user.displayName,
        }));
    }
    /**
     * The current sign-in user (or null)
     * @return {?}
     */
    get user() {
        return this.fireAuth.auth.currentUser;
    }
    /**
     * The path to the profile in firestore
     * @return {?}
     */
    get path() {
        return this.constructor['path'] || this.collectionPath;
    }
    /**
     * Start listening on User
     * @return {?}
     */
    sync() {
        return this.fireAuth.authState.pipe(switchMap((/**
         * @param {?} user
         * @return {?}
         */
        (user) => user ? combineLatest([
            of(user.uid),
            this.selectProfile(user),
            this.selectRoles(user),
        ]) : of([null, null, null]))), tap((/**
         * @param {?} __0
         * @return {?}
         */
        ([uid, profile, roles]) => this.store.update((/** @type {?} */ ({ uid, profile, roles }))))));
    }
    /**
     * \@description Delete user from authentication service and database
     * WARNING This is security sensitive operation
     * @return {?}
     */
    delete() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.user) {
                throw new Error('No user connected');
            }
            /** @type {?} */
            const batch = this.db.firestore.batch();
            const { ref } = this.collection.doc(this.user.uid);
            batch.delete(ref);
            if (this.onDelete) {
                yield this.onDelete(batch);
            }
            yield batch.commit();
            return this.user.delete();
        });
    }
    /**
     * Update the current profile of the authenticated user
     * @param {?} profile
     * @return {?}
     */
    update(profile) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.user.uid) {
                throw new Error('No user connected.');
            }
            /** @type {?} */
            const batch = this.db.firestore.batch();
            const { ref } = this.collection.doc(this.user.uid);
            batch.update(ref, profile);
            if (this.onCreate) {
                yield this.onCreate(profile, batch);
            }
            return batch.commit();
        });
    }
    /**
     * Create a user based on email and password
     * @param {?} email
     * @param {?} password
     * @return {?}
     */
    signup(email, password) {
        return __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            const cred = yield this.fireAuth.auth.createUserWithEmailAndPassword(email, password);
            /** @type {?} */
            const profile = yield this.createProfile(cred.user);
            /** @type {?} */
            const batch = this.db.firestore.batch();
            const { ref } = this.collection.doc(cred.user.uid);
            batch.set(ref, profile);
            if (this.onCreate) {
                this.onCreate(profile, batch);
            }
            return cred;
        });
    }
    /**
     * @param {?=} provider
     * @param {?=} password
     * @return {?}
     */
    signin(provider, password) {
        return __awaiter(this, void 0, void 0, function* () {
            this.store.setLoading(true);
            try {
                /** @type {?} */
                let cred;
                if (!provider) {
                    cred = yield this.fireAuth.auth.signInAnonymously();
                }
                else if (password) {
                    cred = yield this.fireAuth.auth.signInWithEmailAndPassword(provider, password);
                }
                else if (isFireAuthProvider(provider)) {
                    /** @type {?} */
                    const authProvider = getAuthProvider(provider);
                    cred = yield this.fireAuth.auth.signInWithPopup(authProvider);
                }
                else {
                    cred = yield this.fireAuth.auth.signInWithCustomToken(provider);
                }
                if (cred.additionalUserInfo.isNewUser) {
                    /** @type {?} */
                    const profile = yield this.createProfile(cred.user);
                    /** @type {?} */
                    const batch = this.db.firestore.batch();
                    const { ref } = this.collection.doc(cred.user.uid);
                    batch.set(ref, profile);
                    if (this.onCreate) {
                        yield this.onCreate(profile, batch);
                    }
                    batch.commit();
                }
                this.store.setLoading(false);
                return cred;
            }
            catch (err) {
                this.store.setLoading(false);
                if (err.code === 'auth/operation-not-allowed') {
                    console.warn('You tried to connect with unenabled auth provider. Enable it in Firebase console');
                }
                throw err;
            }
        });
    }
    /**
     * Signs out the current user
     * @return {?}
     */
    signOut() {
        return this.fireAuth.auth.signOut();
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const initialAuthState = {
    uid: null,
    profile: null,
    loading: false
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} write
 * @param {?} roles
 * @return {?}
 */
function canWrite(write, roles) {
    return roles.some((/**
     * @param {?} role
     * @return {?}
     */
    role => role === write || role === 'write'));
}
/**
 * @param {?} read
 * @param {?} roles
 * @return {?}
 */
function canRead(read, roles) {
    return roles.some((/**
     * @param {?} role
     * @return {?}
     */
    role => role === read || role === 'write'));
}
/**
 * @param {?} role
 * @param {?} roles
 * @return {?}
 */
function hasRole(role, roles) {
    switch (role) {
        case 'write': return roles.includes('write');
        case 'create': return canWrite('create', roles);
        case 'delete': return canWrite('delete', roles);
        case 'update': return canWrite('update', roles);
        case 'read': return roles.includes('read');
        case 'get': return canRead('get', roles);
        case 'list': return canRead('list', roles);
        default: return false;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} __0
 * @return {?}
 */
function shoudCancel({ validate, cancel }) {
    return race([validate.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        _ => false))), cancel.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        _ => true)))]);
}
/**
 * @param {?} __0
 * @return {?}
 */
function waitForCancel({ startWith, endWith, shouldValidate, shouldCancel }) {
    return __awaiter(this, void 0, void 0, function* () {
        startWith();
        /** @type {?} */
        const cancelled = yield race([
            shouldValidate.pipe(map((/**
             * @param {?} _
             * @return {?}
             */
            _ => false))),
            shouldCancel.pipe(map((/**
             * @param {?} _
             * @return {?}
             */
            _ => true)))
        ]).toPromise();
        endWith(cancelled);
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template E, K
 * @param {?} query
 * @param {?} parent
 * @return {?}
 */
function getSubQuery(query, parent) {
    if (typeof query !== 'function') {
        return query;
    }
    return query(parent);
}
/**
 * @param {?} path
 * @return {?}
 */
function isDocPath(path) {
    return path.split('/').length % 2 === 0;
}
/**
 * Transform a path into a collection Query
 * @template T
 * @param {?} path
 * @param {?=} queryFn
 * @return {?}
 */
function collection(path, queryFn) {
    return (/** @type {?} */ ({ path, queryFn }));
}
/**
 * Transform a path into a doc query
 * @template T
 * @param {?} path
 * @return {?}
 */
function doc(path) {
    return (/** @type {?} */ ({ path }));
}
/**
 * Check if a value is a query
 * @template T
 * @param {?} query
 * @return {?}
 */
function isQuery(query) {
    if (typeof query === 'object') {
        return !!query['path'];
    }
    return false;
}
/**
 * Sync the collection
 * @template E
 * @this {?}
 * @param {?} query The query to trigger
 * @return {?}
 */
function syncQuery(query) {
    // If single query
    if (typeof query === 'string') {
        return isDocPath(query)
            ? this.syncDoc({ path: query })
            : this.syncCollection(query);
    }
    if (Array.isArray(query)) {
        return combineLatest(query.map((/**
         * @param {?} oneQuery
         * @return {?}
         */
        oneQuery => syncQuery.bind(this, oneQuery))));
    }
    if (!isQuery(query)) {
        throw new Error('Query should be either a path, a Query object or an array of Queries');
    }
    ////////////////
    // SUBQUERIES //
    ////////////////
    /**
     * Listen on Child actions
     * @type {?}
     */
    const fromChildAction = (/**
     * @param {?} actions
     * @param {?} child
     * @return {?}
     */
    (actions, child) => {
        /** @type {?} */
        const idKey = 'id';
        // TODO: Improve how to
        const { parentId, key } = child;
        for (const action of actions) {
            /** @type {?} */
            const id = action.payload.doc.id;
            /** @type {?} */
            const data = action.payload.doc.data();
            switch (action.type) {
                case 'added': {
                    this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    (entity) => (/** @type {?} */ (({
                        [key]: arrayUpsert((/** @type {?} */ (entity[key])), id, data, idKey)
                    })))));
                    break;
                }
                case 'removed': {
                    this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    (entity) => (/** @type {?} */ (({
                        [key]: arrayRemove((/** @type {?} */ (entity[key])), id, idKey)
                    })))));
                    break;
                }
                case 'modified': {
                    this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    (entity) => (/** @type {?} */ (({
                        [key]: arrayUpdate((/** @type {?} */ (entity[key])), id, data, idKey)
                    })))));
                }
            }
        }
    });
    /**
     * Stay in sync a subquery of the collection
     * \@param subQuery The sub query based on the entity
     * \@param child An object that describe the parent
     * @type {?}
     */
    const syncSubQuery = (/**
     * @template K
     * @param {?} subQuery
     * @param {?} child
     * @return {?}
     */
    (subQuery, child) => {
        const { parentId, key } = child;
        // If it's a static value
        if (!isQuery(subQuery)) {
            /** @type {?} */
            const update = this['store'].update((/** @type {?} */ (parentId)), (/** @type {?} */ ({ [key]: subQuery })));
            return of(update);
        }
        if (Array.isArray(subQuery)) {
            /** @type {?} */
            const syncQueries = subQuery.map((/**
             * @param {?} oneQuery
             * @return {?}
             */
            oneQuery => {
                if (isQuery(subQuery)) {
                    /** @type {?} */
                    const id = getIdAndPath({ path: subQuery.path });
                    return this['db'].doc(subQuery.path).valueChanges().pipe(tap((/**
                     * @param {?} childDoc
                     * @return {?}
                     */
                    (childDoc) => {
                        this['store'].update((/** @type {?} */ (parentId)), (/**
                         * @param {?} entity
                         * @return {?}
                         */
                        (entity) => (/** @type {?} */ (({
                            [key]: arrayAdd((/** @type {?} */ (entity[key])), id, childDoc)
                        })))));
                    })));
                }
                return syncSubQuery(oneQuery, child);
            }));
            return combineLatest(syncQueries);
        }
        if (typeof subQuery !== 'object') {
            throw new Error('Query should be either a path, a Query object or an array of Queries');
        }
        // Sync subquery
        if (isDocPath(subQuery.path)) {
            return this['db'].doc(subQuery.path).valueChanges().pipe(tap((/**
             * @param {?} children
             * @return {?}
             */
            (children) => this['store'].update((/** @type {?} */ (parentId)), (/** @type {?} */ ({ [key]: children }))))));
        }
        else {
            return (/** @type {?} */ (this['db'].collection(subQuery.path, subQuery.queryFn)
                .stateChanges()
                .pipe(withTransaction((/**
             * @param {?} actions
             * @return {?}
             */
            actions => fromChildAction(actions, child))))));
        }
    });
    /**
     * Get in sync with all the subqueries
     * \@param subQueries A map of all the subqueries
     * \@param parent The parent entity to attach the children to
     * @type {?}
     */
    const syncAllSubQueries = (/**
     * @param {?} subQueries
     * @param {?} parent
     * @return {?}
     */
    (subQueries, parent) => {
        /** @type {?} */
        const obs = Object.keys(subQueries)
            .filter((/**
         * @param {?} key
         * @return {?}
         */
        key => key !== 'path' && key !== 'queryFn'))
            .map((/**
         * @param {?} key
         * @return {?}
         */
        (key) => {
            /** @type {?} */
            const queryLike = getSubQuery(subQueries[key], parent);
            /** @type {?} */
            const child = { key, parentId: parent[this.idKey] };
            return syncSubQuery(queryLike, child);
        }));
        return combineLatest(obs);
    });
    ////////////////
    // MAIN QUERY //
    ////////////////
    /**
     * Listen on action with child queries
     * @type {?}
     */
    const fromActionWithChild = (/**
     * @param {?} actions
     * @param {?} mainQuery
     * @param {?} subscriptions
     * @return {?}
     */
    (actions, mainQuery, subscriptions) => {
        for (const action of actions) {
            /** @type {?} */
            const id = action.payload.doc.id;
            /** @type {?} */
            const data = action.payload.doc.data();
            switch (action.type) {
                case 'added': {
                    /** @type {?} */
                    const entity = Object.assign({ [this.idKey]: id }, data);
                    this['store'].upsert(id, entity);
                    subscriptions[id] = syncAllSubQueries((/** @type {?} */ (mainQuery)), entity).subscribe();
                    break;
                }
                case 'removed': {
                    this['store'].remove(id);
                    subscriptions[id].unsubscribe();
                    delete subscriptions[id];
                    break;
                }
                case 'modified': {
                    this['store'].update(id, data);
                }
            }
        }
    });
    const { path, queryFn } = query;
    if (isDocPath(path)) {
        const { id } = this['getIdAndPath']({ path });
        /** @type {?} */
        let subscription;
        return this['db'].doc(path).valueChanges().pipe(tap((/**
         * @param {?} entity
         * @return {?}
         */
        entity => {
            this['store'].upsert(id, (/** @type {?} */ (Object.assign({ id }, entity))));
            if (!subscription) { // Subscribe only the first time
                subscription = syncAllSubQueries((/** @type {?} */ (query)), entity).subscribe();
            }
        })), 
        // Stop subscription
        finalize((/**
         * @return {?}
         */
        () => subscription.unsubscribe())));
    }
    else {
        /** @type {?} */
        const subscriptions = {};
        return (/** @type {?} */ (this['db'].collection(path, queryFn)
            .stateChanges()
            .pipe(withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        actions => fromActionWithChild(actions, query, subscriptions))), 
        // Stop all subscriptions
        finalize((/**
         * @return {?}
         */
        () => Object.keys(subscriptions).forEach((/**
         * @param {?} id
         * @return {?}
         */
        id => {
            subscriptions[id].unsubscribe();
            delete subscriptions[id];
        })))))));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Redirect the
 * @template E
 * @param {?} redirectTo Route path to redirecto if collection is empty
 * @return {?}
 */
function redirectIfEmpty(redirectTo) {
    return map((/**
     * @param {?} actions
     * @return {?}
     */
    (actions) => actions.length === 0 ? redirectTo : true));
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { CollectionConfig, CollectionGroupService, CollectionGuard, CollectionGuardConfig, CollectionService, FireAuthService, SubcollectionService, canRead, canWrite, collection, doc, fireAuthProviders, getIdAndPath, getPathParams, hasRole, initialAuthState, isDocPath, pathWithParams, redirectIfEmpty, shoudCancel, syncFromAction, syncQuery, waitForCancel };
//# sourceMappingURL=akita-ng-fire.js.map
